/* $Header: /home/yav/fkiss/RCS/dir.c,v 1.10 1995/11/28 16:47:35 yav Exp $
 * Simple directory access routine
 * written by yav (UHD98984@pcvan.or.jp)
 *
 * #define USE_BSD_DIRLIB 1
 *  use opendir, readdir and closedir library routines.
 * #define USE_BSD_DIRLIB 0
 *  use popen to execute "ls -f".
 *
 * DIR_STANDALONE_TEST for self checking
 *
 */

#include <stdio.h>

#ifndef DIR_STANDALONE_TEST
# include <X11/Xlib.h>
# include "config.h"
# include "headers.h"
# include "fkiss.h"
# define PUBLIC_DIR_C
# include "extern.h"
#endif

#ifndef USE_BSD_DIRLIB
# if HAVE_SYS_DIR_H
#  if HAVE_OPENDIR
#   if HAVE_READDIR
#    if HAVE_CLOSEDIR
#     define USE_BSD_DIRLIB 1
#    endif
#   endif
#  endif
# endif
#endif

#if USE_BSD_DIRLIB
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
#endif

#ifdef DIR_STANDALONE_TEST
#define ks_malloc malloc
#define ks_realloc realloc
#define ks_strdup strdup
#else
#endif

char id_dir[] = "$Id: dir.c,v 1.10 1995/11/28 16:47:35 yav Exp $";

/* dir_ls - get filename list in directory
 * return filename pointer list (NULL : directory open error)
 */
char **dir_ls(path)
     char *path;		/* directory */
{
  int r;
  char **list;
#if USE_BSD_DIRLIB
  DIR *dp;
  struct direct *p;
#else
  FILE *fp;
  char buf[1024];
#endif
  
  r = 0;
  list = NULL;
#if USE_BSD_DIRLIB
  if ((dp = opendir(path)) != NULL) {
    list = (char **)ks_malloc(sizeof(char *));
    while ((p = readdir(dp)) != NULL) {
      *(list+r++) = ks_strdup(p->d_name);
      list = (char **)ks_realloc(list, sizeof(char *) * (r+1));
    }
    *(list+r) = NULL;
    closedir(dp);
  }
#else /* USE_BSD_DIRLIB */
  strcpy(buf, "ls -f ");
  strcat(buf, path);
  if ((fp = popen(buf, "r")) != NULL) {
    list = (char **)ks_malloc(sizeof(char *));
    while (fgets(buf, sizeof(buf), fp) != NULL) {
      buf[strlen(buf) - 1] = '\0'; /* cut '\n' */
      *(list+r++) = ks_strdup(buf);
      list = (char **)ks_realloc(list, sizeof(char *) * (r+1));
    }
    *(list+r) = NULL;
    pclose(fp);
  }
#endif /* USE_BSD_DIRLIB */
  return list;
}

/* free all allocated filename list */
void dir_free(list)
     char **list;
{
  char **p;
  
  if (list != NULL) {
    for (p = list; *p != NULL; p++)
      free(*p);
    free(list);
  }    
}

#ifdef DIR_STANDALONE_TEST
void main(argc, argv)
     int argc;
     char **argv;
{
  int i, n;
  char **buf;
  
  if (argc < 2)
    exit(1);
#if USE_BSD_DIRLIB
  printf("* dir lib *\n");
#else
  printf("* dir ls  *\n");
#endif
  buf = dir_ls(*++argv);
  if (buf == NULL) {
    fprintf(stderr, "``%s'' read error!\n", *argv);
    exit(1);
  }
  for (i = 0; *(buf+i) != NULL; i++) {
    printf("[%s]\n", *(buf+i));
  }
  dir_free(buf);
  printf("Total %d files\n", i);
  exit(0);
}
#endif

/* End of file */
