/* $Header: /home/yav/fkiss/RCS/sound.c,v 1.11 1995/12/03 05:07:47 yav Exp $
 * fkiss sound routine for Linux
 * written by yav (UHD98984@pcvan.or.jp)
 */

#include <X11/Xos.h>
#include <stdio.h>
#include "config.h"
#include "headers.h"

#if HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef linux
# define USE_DSP
# define DEV_SOUND	"/dev/dsp"
# include <fcntl.h>
# include <sys/ioctl.h>
# include <linux/soundcard.h>
#endif

#ifdef __FreeBSD__
# define USE_DSP
# define DEV_SOUND	"/dev/dsp"
# include <machine/soundcard.h>
#endif

#ifndef DEV_SOUND
# define DEV_SOUND	"/dev/audio"
#endif

char id_sound[] = "$Id: sound.c,v 1.11 1995/12/03 05:07:47 yav Exp $";
char *sound_device = DEV_SOUND;
int sound_debug = 0;		/* bit0: print debug information
				 * bit1: not output to device
				 */
static int sound_enable = 0;	/* 0:disable other:enable */

int sound_rate = 22050;		/* sampling rate Hz */
int sound_channels = 1;		/* 1:mono 2:stereo */
int sound_bits = 8;		/* bits per sample 8 or 16 */
int sound_ulaw_encoded = 0;	/* input ulaw encoded */
int sound_little_endian = 0;	/* input little-endian */
int sound_unsigned = -1;	/* input zero level 0x80 */
int sound_force = 0;		/* no conversion, force write DSP data */

#define SOUND_CACHE_MAX 64
static int cache_n;
typedef struct {
  char *pathname;
  char *buffer;
  long size;
} CACHE_TABLE;
static CACHE_TABLE *cache_list = NULL;

#ifdef USE_DSP

static int sample_bytes;	/* byts per sample 1 or 2 */
static int dsp_rate;
static int dsp_bits;
static int dsp_channels;

#include "ulaw.h"

#endif

void adjust_parms()
{
  if (sound_bits < 1)
    sound_bits = 8;
#ifdef USE_DSP
  sample_bytes = (sound_bits+7) / 8;
  /* defaut 8bit data zero level 0x80 */
  if (sound_unsigned < 0)
    sound_unsigned = (sample_bytes == 1);
#endif
}

#ifdef USE_DSP

int setup_dsp(fd)
     int fd;
{
  /* copy to DSP setting parms */
  dsp_channels = sound_channels;
  dsp_bits = sound_bits;
  dsp_rate = sound_rate;
  ioctl(fd, SOUND_PCM_WRITE_CHANNELS, &dsp_channels);
  ioctl(fd, SOUND_PCM_WRITE_BITS, &dsp_bits);
  ioctl(fd, SOUND_PCM_WRITE_RATE, &dsp_rate);
  return 0;
}

void conv_center80(buf, n)
     unsigned char *buf;
     int n;
{
  while (n--) {
    *buf += 0x80;
    buf++;
  }
}

/* convert ulaw to  0x80 center 8bit linear PCM */
void conv_ulaw_dsp(buf, n)
     unsigned char *buf;
     int n;
{
  while (n--)
    *buf++ = ulaw_dsp[*buf];
}

#endif /* USE_DSP */

int play_sound(fd, buf, n)
     int fd;
     char *buf;
     int n;
{
#ifdef USE_DSP
  int i, d;
  unsigned char *p;
  unsigned char *p0;
  
  if (!sound_force) {
    if (sound_ulaw_encoded) {
      conv_ulaw_dsp(buf, n);
    } else {
      if (sample_bytes != 1 || sound_channels != 1) {
	/* pickup highest byte */
	d = sample_bytes * sound_channels;
	p0 = p = buf;
	if (sound_little_endian)
	  p0 += sample_bytes-1;
	if (sound_channels == 1) {
	  for (i = 0; i < n; i += d) {
	    *p++ = *p0;
	    p0 += d;
	  }
	} else {
	  for (i = 0; i < n; i += d) {
	    *p++ = *p0;
	    *p++ = *(p0+sample_bytes);
	    p0 += d;
	  }
	}
	n /= sample_bytes;
      }
      /* adjust zero level 0x80 */
      if (!sound_unsigned)
	conv_center80(buf, n);
    }
  }
#endif
  write(fd, buf, n);
  return 0;
}

long get_long(p)
     unsigned char *p;
{
  return sound_little_endian ?
    (((long)*(p+3)) << 24) | (((long)*(p+2)) << 16) | (*(p+1) << 8) | *p :
      (((long)*p) << 24) | (((long)*(p+1)) << 16) | (*(p+2) << 8) | *(p+3);
}

/* check Sun Audio file header
 * return 0 - yes Sun Audio file.
 *        other - byte count in buf to play
 *   fp == NULL
 *        return data offset to play
 */
int check_au_header(fp, buf)
     FILE *fp;			/* == NULL -> data already loaded */
     char *buf;
{
  int i;
  long data_offset;
  long data_size;
  long data_type;
  long data_rate;
  long data_channel;
  static struct {
    int bits;			/* bits per sample */
    int ulaw;			/* 0:linear 1:ulaw encoded */
  } type_info[] = {
    { 0, 0},			/* 0: dummy */
    { 8, 1},			/* 1: 8-bit ISDN u-law */
    { 8, 0},			/* 2: 8-bit linear PCM */
    {16, 0},			/* 3:16-bit linear PCM */
    {24, 0},			/* 4:24-bit linear PCM */
    {32, 0},			/* 5:32-bit linear PCM */
  };
  
  if (fp != NULL) {
    i = fread(buf, 1, 8, fp);
    if (i != 8)
      return i;
  }
  if (strncmp(buf, ".snd", 4) != 0) {
    if (strncmp(buf, "dns.", 4) != 0)
      return fp == NULL ? 0 : 8;
    sound_little_endian = 1;
  }
  if (fp != NULL)
    if (!fread(buf+8, 0x18-8, 1, fp))
      return 8;
  data_offset  = get_long(buf+0x04);
  data_size    = get_long(buf+0x08);
  data_type    = get_long(buf+0x0c);
  data_rate    = get_long(buf+0x10);
  data_channel = get_long(buf+0x14);
  if (fp != NULL) {
    i = data_offset - 0x18;
    while (i--)
      getc(fp);
  }
  if (data_type < 1 || data_type > 5)
    return fp == NULL ? 0 : 0x18;
  sound_bits = type_info[data_type].bits;
  sound_ulaw_encoded = type_info[data_type].ulaw;
  sound_rate = data_rate;
  sound_channels = data_channel;
  return fp == NULL ? data_offset : 0;
}

void disp_sound_info(name)
     char *name;
{
  fprintf(stderr, "device: %s\n", sound_device);
  fprintf(stderr, "input: %s\n", name == NULL ? "(stdin)" : name);
  fprintf(stderr, " %5dHz %2d-bit %d-channel",
	  sound_rate, sound_bits, sound_channels);
  if (sound_ulaw_encoded)
    fprintf(stderr, " u-law");
  if (sound_little_endian)
    fprintf(stderr, " Little-endian");
  if (!sound_ulaw_encoded)
    fprintf(stderr, " %s", sound_unsigned ? "unsigned" : "signed");
  fprintf(stderr, "\n");
}

#ifdef USE_DSP
void disp_dsp_info()
{
  fprintf(stderr, " %5dHz %2d-bit %d-channel\n",
	  dsp_rate, dsp_bits, dsp_channels);
}

#endif /* USE_DSP */

/* Check sound device
 * return 0:OK other:NG
 */
int sound_init()
{
  int fd;
  struct stat st;
  
  cache_n = 0;
  /* sound device exist? */
  if (stat(sound_device, &st) == 0) {
    /* sound device writeble? */
    fd = open(sound_device, O_WRONLY);
    if (fd >= 0) {
      close(fd);
      sound_enable = 1;
      cache_list = malloc(sizeof(CACHE_TABLE)*SOUND_CACHE_MAX);
    }
  }
  return !sound_enable;
}

void sound_end()
{
  int i;
  
  for (i = 0; i < cache_n; i++)
    free((cache_list+i)->buffer);
  if (cache_list != NULL)
    free(cache_list);
}

/* return audio file size
 *  minus : file not found
 */
long sound_filesize(name)
     char *name;
{
  long r;
  struct stat st;
  
  r = -1;
  if (sound_enable && stat(name, &st) == 0)
    r = st.st_size;
  return r;
}

/* cache audio data
 * return minus cannot to cache
 */
int sound_cache(name)
     char *name;
{
  char *p;
  FILE *fp;
  CACHE_TABLE *cp;
  
  if (!sound_enable)
    return -1;
  if (cache_n >= SOUND_CACHE_MAX)
    return -1;
  cp = cache_list + cache_n;
  cp->size = sound_filesize(name);
  if (cp->size < 0)
    return -1;
  fp = fopen(name, "r");
  if (fp == NULL)
    return -1;			/* Why? stat success, but fopen fail??? */
  /* If name will be free, use strdup to keep string */
  cp->pathname = name;
  p = malloc(cp->size);
  if (p == NULL) {
    fclose(fp);
    return -1;
  }
  fread(p, cp->size, 1, fp);
  cp->buffer = p;
  cache_n++;
  fclose(fp);
  return 0;
}

int cached_no(name)
     char *name;
{
  int i;
  
  for (i = 0; i < cache_n; i++)
    if (strcmp(name, (cache_list+i)->pathname) == 0)
      return i;
  return -1;			/* not cached */
}

/* Play audio file
 * return 0: No error 1:file open error 2:device open error
 */
int sound_play(name)
     char *name;		/* audio filename */
{
  int i, fd;
  int cn;
  FILE *fp;
  char buf[32*1024];		/* 8 * N bytes */
  
  if (sound_enable) {
    if (name == NULL) {
      fp = stdin;
    } else {
      cn = cached_no(name);
      if (cn >= 0) {
	fp = NULL;
      } else {
	fp = fopen(name, "r");
	if (fp == NULL)
	  return 1;		/* Sound file open error */
      }
    }
    i = check_au_header(fp, buf);
    adjust_parms();
    if (sound_debug & 1)
      disp_sound_info(name);
    fd = open(sound_device, O_WRONLY);
    if (fd < 0) {
      fclose(fp);
      return 2;			/* device open error */
    }
#ifdef USE_DSP
    setup_dsp(fd);
    if (sound_debug & 1)
      disp_dsp_info();
#endif
    if (!(sound_debug & 2)) {
      if (fp == NULL) {
	play_sound(fd, (cache_list+cn)->buffer+i, (cache_list+cn)->size-i);
      } else {
	do {
	  if (i)
	    play_sound(fd, buf, i);
	} while ((i = fread(buf, 1, sizeof(buf), fp)) != 0);
      }
    }
    close(fd);
    fclose(fp);
  }
  return 0;
}

/* End of file */
